@@ -45,7 +45,7 @@ class Namespace(object):
     if not(base_iri.endswith("#") or base_iri.endswith("/") or base_iri.endswith(":")): raise ValueError("base_iri must end with '#', '/' or ':' !")
     name = name or base_iri[:-1].rsplit("/", 1)[-1]
     if name.endswith(".owl") or name.endswith(".rdf"): name = name[:-4]
-    
+
     if   isinstance(world_or_ontology, Ontology):
       self.ontology = world_or_ontology
       self.world    = world_or_ontology.world
@@ -55,32 +55,48 @@ class Namespace(object):
       self.ontology = None
       self.world    = world_or_ontology
       self.world._namespaces[base_iri] = self
-      
     else:
       self.ontology = None
       self.world    = None
       
-    self.base_iri = base_iri
-    self.name     = name
+    self._base_iri = base_iri
+    self.name      = name
     
+  def get_base_iri(self): return self._base_iri
+  # def set_base_iri(self, new_base_iri, rename_entities = True):
+  #   if rename_entities == False: raise ValueError("set_base_iri() with rename_entity=False is ontly supported on Ontology, not on Namespace. Please create a new Namespace.")
+  #   if self.world.graph: self.world.graph.acquire_write_lock()
+    
+  #   if self.ontology: del self.ontology._namespaces[self._base_iri]
+  #   else:             del self.world   ._namespaces[self._base_iri]
+    
+  #   self.world._refactor_onto(self.storid, self._base_iri, new_base_iri)
+      
+  #   self._base_iri = new_base_iri
+  #   if self.ontology: self.ontology._namespaces[new_base_iri] = self
+  #   else:             self.world   ._namespaces[new_base_iri] = self
+    
+  #   if self.world.graph: self.world.graph.release_write_lock()
+  base_iri = property(get_base_iri)#, set_base_iri)
+  
   def __enter__(self):
     if self.ontology is None: raise ValueError("Cannot assert facts in this namespace: it is not linked to an ontology! (it is probably a global namespace created by get_namespace(); please use your_ontology.get_namespace() instead)")
-    if self.world.graph: self.world.graph.acquire_write_lock()
-    #CURRENT_NAMESPACES.append(self)
+    if self.world.graph:
+      self.world.graph.acquire_write_lock()
     l = CURRENT_NAMESPACES.get()
     if l is None: CURRENT_NAMESPACES.set([self])
     else:         l.append(self)
     return self
-    
-  def __exit__(self, exc_type = None, exc_val = None, exc_tb = None):
-    #del CURRENT_NAMESPACES[-1]
-    del CURRENT_NAMESPACES.get()[-1]
-    if self.world.graph: self.world.graph.release_write_lock()
-    
-  def __repr__(self): return """%s.get_namespace("%s")""" % (self.ontology, self.base_iri)
   
-  def __getattr__(self, attr): return self.world["%s%s" % (self.base_iri, attr)] #return self[attr]
-  def __getitem__(self, name): return self.world["%s%s" % (self.base_iri, name)]
+  def __exit__(self, exc_type = None, exc_val = None, exc_tb = None):
+    del CURRENT_NAMESPACES.get()[-1]
+    if self.world.graph:
+      self.world.graph.release_write_lock()
+      
+  def __repr__(self): return """%s.get_namespace("%s")""" % (self.ontology, self._base_iri)
+  
+  def __getattr__(self, attr): return self.world["%s%s" % (self._base_iri, attr)] #return self[attr]
+  def __getitem__(self, name): return self.world["%s%s" % (self._base_iri, name)]
 
 class _GraphManager(object):
   def _abbreviate  (self, iri, create_if_missing = True):
@@ -253,14 +269,16 @@ WHERE q1.p=? AND q1.o=?
     l = []
     while bnode and (bnode != rdf_nil):
       first, d = self._get_triple_sp_od(bnode, rdf_first)
-      if first != rdf_nil: l.append(self._to_python(first, d))
+      if not ((first == rdf_nil) and (d is None)): l.append(self._to_python(first, d))
       bnode = self._get_obj_triple_sp_o(bnode, rdf_rest)
     return l
   
   def _parse_list_as_rdf(self, bnode):
     while bnode and (bnode != rdf_nil):
       first, d = self._get_triple_sp_od(bnode, rdf_first)
-      if first != rdf_nil: yield first, d
+      if not ((first == rdf_nil) and (d is None)): yield first, d
+      #if (first == rdf_nil) and (d is None): return
+      #yield first, d
       bnode = self._get_obj_triple_sp_o(bnode, rdf_rest)
       
   def _to_python(self, o, d = None, main_type = None, main_onto = None, default_to_none = False):
@@ -427,15 +445,16 @@ class World(_GraphManager):
   def __init__(self, backend = "sqlite", filename = ":memory:", dbname = "owlready2_quadstore", **kargs):
     global owl_world
     
-    self.world            = self
-    self.filename         = filename
-    self.ontologies       = {}
-    self._props           = {}
-    self._reasoning_props = {}
-    self._entities        = weakref.WeakValueDictionary()
-    self._namespaces      = weakref.WeakValueDictionary()
-    self._rdflib_store    = None
-    self.graph            = None
+    self.world               = self
+    self.filename            = filename
+    self.ontologies          = {}
+    self._props              = {}
+    self._reasoning_props    = {}
+    self._entities           = weakref.WeakValueDictionary()
+    self._namespaces         = weakref.WeakValueDictionary()
+    self._fusion_class_cache = {}
+    self._rdflib_store       = None
+    self.graph               = None
     
     if not owl_world is None:
       self._entities.update(owl_world._entities) # add OWL entities in the world
@@ -445,10 +464,12 @@ class World(_GraphManager):
     if filename:
       self.set_backend(backend, filename, dbname, **kargs)
       
+    self.get_ontology("http://anonymous/") # Pre-create, in order to avoird creation during a reading sequence
+    
   def set_backend(self, backend = "sqlite", filename = ":memory:", dbname = "owlready2_quadstore", **kargs):
     if   backend == "sqlite":
       from owlready2.triplelite import Graph
-      if self.graph and len(self.graph):
+      if self.graph and (len(self.graph) > 1): # 1 is for http://anonymous ontology
         self.graph = Graph(filename, world = self, clone = self.graph, **kargs)
       else:
         self.graph = Graph(filename, world = self, **kargs)
@@ -469,8 +490,21 @@ class World(_GraphManager):
 
     self._full_text_search_properties = CallbackList([self._get_by_storid(storid, default_to_none = True) or storid for storid in self.graph.get_fts_prop_storid()], self, World._full_text_search_changed)
     
-  def close(self): self.graph.close()
-  
+  def close(self):
+    self._destroy_cached_entities()
+    self.graph.close()
+    
+  def _destroy_cached_entities(self):
+    _entities = self._entities
+    _fusion_classes = set(self._fusion_class_cache.values())
+    for i, cached in enumerate(_cache):
+      if (not cached is None) and ((cached.namespace.world is self) or (cached in _fusion_classes)):
+        _cache[i] = None
+    self._entities.clear()
+
+  def forget_reference(self, python_entity):
+    self._entities.pop(python_entity.storid, None)
+    
   def get_full_text_search_properties(self): return self._full_text_search_properties
   def set_full_text_search_properties(self, l):
     old = self._full_text_search_properties
@@ -521,23 +555,23 @@ class World(_GraphManager):
   def prepare_sparql(self, sparql, error_on_undefined_entities = True): # lru_cache does not handle optional args
     return self._prepare_sparql(sparql, error_on_undefined_entities)
   
-  def get_ontology(self, base_iri):
+  def get_ontology(self, base_iri, OntologyClass = None):
     if (not base_iri.endswith("/")) and (not base_iri.endswith("#")):
       if   ("%s/" % base_iri) in PREDEFINED_ONTOLOGIES: base_iri = base_iri = "%s/" % base_iri
       elif ("%s#" % base_iri) in self.ontologies:       base_iri = base_iri = "%s#" % base_iri
       elif ("%s/" % base_iri) in self.ontologies:       base_iri = base_iri = "%s/" % base_iri
       else:                                             base_iri = base_iri = "%s#" % base_iri
     if base_iri in self.ontologies: return self.ontologies[base_iri]
-    return Ontology(self, base_iri)
-  
-  def get_namespace(self, base_iri, name = ""):
+    return (OntologyClass or Ontology)(self, base_iri)
+    
+  def get_namespace(self, base_iri, name = "", NamespaceClass = None):
     if (not base_iri.endswith("/")) and (not base_iri.endswith("#")) and (not base_iri.endswith(":")):
       if   ("%s#" % base_iri) in self.ontologies: base_iri = base_iri = "%s#" % base_iri
       elif ("%s/" % base_iri) in self.ontologies: base_iri = base_iri = "%s/" % base_iri
       elif ("%s:" % base_iri) in self.ontologies: base_iri = base_iri = "%s:" % base_iri
       else:                                       base_iri = base_iri = "%s#" % base_iri
     if base_iri in self._namespaces: return self._namespaces[base_iri]
-    return Namespace(self, base_iri, name or base_iri[:-1].rsplit("/", 1)[-1])
+    return (NamespaceClass or Namespace)(self, base_iri, name or base_iri[:-1].rsplit("/", 1)[-1])
     
   def _del_triple_with_update(self, s, p, o, d = None):
     sub = None
@@ -582,7 +616,7 @@ class World(_GraphManager):
     if d is None: self._del_obj_triple_raw_spo  (s,p,o)
     else:         self._del_data_triple_raw_spod(s,p,o,d)
     
-  def _add_triple_with_update(self, ontology0, s, p, o, d = None):
+  def _add_triples_with_update(self, ontology0, triples):
     l = owlready2.namespace.CURRENT_NAMESPACES.get()
     if l:
       ontology = l[-1].ontology
@@ -591,45 +625,69 @@ class World(_GraphManager):
       if not ontology:
         raise ValueError("Cannot add triples outside a 'with' block. Please start a 'with' block to indicate in which ontology the new triple is added, or include a 'WITH <onto_IRI>' statement in SPARQL.")
       
-    sub = None
-    if   (s > 0) and (s in self.world._entities): sub = self._entities[s]
-    elif (s < 0) and (s in ontology._bnodes):     sub = ontology._bnodes[s]
-    if not sub is None:
-      prop = self._entities.get(p)
-      if   prop:
-        try: delattr(sub, prop.python_name)
-        except: pass
-        
-      elif d is None:
-        if o > 0: obj = self._load_by_storid(o)
-        else:     obj = self._parse_bnode(o)
-        
-        if not obj is None:
-          if   (p == rdf_type) or (p == rdfs_subclassof) or (p == rdfs_subpropertyof):
-            with DONT_COPY_BN:
-              with ontology: sub.is_a.append(obj)
-            return
-          
-          elif (p == owl_equivalentindividual) or (p == owl_equivalentclass) or (p == owl_equivalentproperty):
-            with DONT_COPY_BN:
-              with ontology: sub.equivalent_to.append(obj)
-            return
+    is_a_triples = defaultdict(list)
+    
+    for triple in triples:
+      if len(triple) == 3: s, p, o    = triple; d = None
+      else:                s, p, o, d = triple
+      
+      if (p == rdf_type) or (p == rdfs_subclassof):
+        is_a_triples[s, p].append(o)
+        continue
+      
+      if   (s > 0) and (s in self.world._entities): sub = self._entities[s]
+      elif (s < 0) and (s in ontology._bnodes):     sub = ontology._bnodes[s]
+      else:                                         sub = None
+      if not sub is None:
+        prop = self._entities.get(p)
+        if   prop:
+          try: delattr(sub, prop.python_name)
+          except: pass
           
+        elif d is None:
+          if   (p == owl_equivalentindividual) or (p == owl_equivalentclass) or (p == owl_equivalentproperty):
+            obj = self._get_by_storid(o) if o > 0 else self._parse_bnode(o)
+            if not obj is None:
+              with DONT_COPY_BN:
+                with ontology: sub.equivalent_to.append(obj)
+              continue
+            
           elif (p == owl_inverse_property):
-            with ontology: sub.inverse_property = obj
-            return
+            obj = self._get_by_storid(o) if o > 0 else self._parse_bnode(o)
+            if not obj is None:
+              with ontology: sub.inverse_property = obj
+              continue
           
           elif (p == rdf_domain):
-            with ontology: sub.domain.append(obj)
-            return
-          
+            obj = self._get_by_storid(o) if o > 0 else self._parse_bnode(o)
+            if not obj is None:
+              with ontology: sub.domain.append(obj)
+              continue
+            
           elif (p == rdf_range):
-            with ontology: sub.range.append(obj)
-            return
+            obj = self._get_by_storid(o) if o > 0 else self._parse_bnode(o)
+            if not obj is None:
+              with ontology: sub.range.append(obj)
+              continue
+            
+      if d is None: ontology.graph._add_obj_triple_raw_spo  (s, p, o)
+      else:         ontology.graph._add_data_triple_raw_spod(s, p, o, d)
+      
+    # Factorize is_a triples for better performance
+    for (s, p), os in is_a_triples.items():
+      if   (s > 0) and (s in self.world._entities): sub = self._entities[s]
+      elif (s < 0) and (s in ontology._bnodes):     sub = ontology._bnodes[s]
+      else:                                         sub = None
+      if not sub is None:
+        objs = [self._get_by_storid(o) if o > 0 else self._parse_bnode(o) for o in os]
+        objs = [obj for obj in objs if not obj is None]
+        
+        with DONT_COPY_BN:
+          with ontology: sub.is_a.extend(objs)
           
-    if d is None: ontology.graph._add_obj_triple_raw_spo  (s,p,o)
-    else:         ontology.graph._add_data_triple_raw_spod(s,p,o,d)
-    
+      else:
+        for o in os: ontology.graph._add_obj_triple_raw_spo(s, p, o)
+        
     
   def get(self, iri, default = None):
     storid = self._abbreviate(iri, False)
@@ -657,14 +715,16 @@ class World(_GraphManager):
     with LOADING:
       types       = []
       is_a_bnodes = []
+
       for graph, obj in self._get_obj_triples_sp_co(storid, rdf_type):
         if main_onto is None: main_onto = self.graph.context_2_user_context(graph)
-        if   obj == owl_class: main_type = ThingClass
+        if   obj == owl_class:               main_type = ThingClass
         elif obj == owl_object_property:     main_type = ObjectPropertyClass;     types.append(ObjectProperty)
         elif obj == owl_data_property:       main_type = DataPropertyClass;       types.append(DataProperty)
         elif obj == owl_annotation_property: main_type = AnnotationPropertyClass; types.append(AnnotationProperty)
         elif (obj == owl_named_individual) or (obj == owl_thing):
           if main_type is None: main_type = Thing
+        elif 105 <= obj <= 109:              main_type = ObjectPropertyClass;     types.append(ObjectProperty) # TransitiveProperty, SymmetricProperty, AsymmetricProperty, ReflexiveProperty, IrreflexiveProperty
         else:
           if not main_type: main_type = Thing
           if obj < 0: is_a_bnodes.append((self.graph.context_2_user_context(graph), obj))
@@ -695,11 +755,11 @@ class World(_GraphManager):
             obj2 = self._entities.get(obj)
             if obj2 is None: obj2 = self._load_by_storid(obj, None, main_type, main_onto, default_to_none, trace)
             if not obj2 is None: is_a_entities.append(obj2)
-
+            
       if main_onto is None:
         main_onto = self.get_ontology("http://anonymous/")
         full_iri = full_iri or self._unabbreviate(storid)
-        if full_iri.startswith(owl.base_iri) or full_iri.startswith(rdfs.base_iri) or full_iri.startswith("http://www.w3.org/1999/02/22-rdf-syntax-ns#"): return None
+        if full_iri.startswith(owl._base_iri) or full_iri.startswith(rdfs._base_iri) or full_iri.startswith("http://www.w3.org/1999/02/22-rdf-syntax-ns#"): return None
         
       if main_onto:
         if isinstance(storid, int) and (storid < 0):
@@ -792,15 +852,27 @@ class World(_GraphManager):
      
 class Ontology(Namespace, _GraphManager):
   def __init__(self, world, base_iri, name = None):
+    #need_write = False
+    if world.graph: world.graph.acquire_write_lock()
+    
     self.world       = world # Those 2 attributes are required before calling Namespace.__init__
     self._namespaces = weakref.WeakValueDictionary()
     Namespace.__init__(self, self, base_iri, name)
     self.loaded                = False
     self._bnodes               = weakref.WeakValueDictionary()
     self.storid                = world._abbreviate(base_iri[:-1])
+    #self.storid                = world._abbreviate(base_iri[:-1], False)
+    #if self.storid is None:
+    #  if world.graph and not need_write:
+    #    need_write = True
+    #    world.graph.acquire_write_lock()
+    #  self.storid              = world._abbreviate(base_iri[:-1], True)
+      
     self._imported_ontologies  = CallbackList([], self, Ontology._import_changed)
     self.metadata              = Metadata(self, self.storid)
     
+    #if world.graph: world.graph.acquire_write_lock()
+    
     if world.graph is None:
       self.graph = None
     else:
@@ -810,23 +882,81 @@ class Ontology(Namespace, _GraphManager):
       if not new_in_quadstore:
         self._load_properties()
         
-    world.ontologies[self.base_iri] = self
-    if _LOG_LEVEL: print("* Owlready2 * Creating new ontology %s <%s>." % (self.name, self.base_iri), file = sys.stderr)
+    world.ontologies[self._base_iri] = self
+    if _LOG_LEVEL: print("* Owlready2 * Creating new ontology %s <%s>." % (self.name, self._base_iri), file = sys.stderr)
     
     if (not LOADING) and (not self.graph is None):
       if not self._has_obj_triple_spo(self.storid, rdf_type, owl_ontology):
-        if self.world.graph: self.world.graph.acquire_write_lock()
+        #if not need_write:
+        #  need_write = True
+        #  world.graph.acquire_write_lock()
         self._add_obj_triple_spo(self.storid, rdf_type, owl_ontology)
-        if self.world.graph: self.world.graph.release_write_lock()
         
     if not self.world._rdflib_store is None: self.world._rdflib_store._add_onto(self)
     
-  def destroy(self):
+    if world.graph: world.graph.release_write_lock()
+    #if need_write: world.graph.release_write_lock()
+
+  def get_base_iri(self): return self._base_iri
+  def set_base_iri(self, new_base_iri, rename_entities = True):
+    if self.world.graph: self.world.graph.acquire_write_lock()
+
+    del self.world.ontologies[self._base_iri]
+    del self._namespaces[self._base_iri]
+    
+    old_base_iri = self._base_iri
+    if rename_entities:
+      self.world._refactor_onto(self.storid, old_base_iri, new_base_iri)
+    else:
+      self.world._refactor(self.storid, new_base_iri)
+      
+    self._base_iri = new_base_iri
+    self.world.ontologies[new_base_iri] = self._namespaces[new_base_iri] = self
+    
+    if rename_entities: # Update Namespaces with the same base IRI
+      for d, namespace in [(self.world._namespaces, v) for v in self.world._namespaces.values()] + [(ontology._namespaces, v) for ontology in self.world.ontologies.values() for v in ontology._namespaces.values()]:
+        if (not isinstance(namespace, Ontology)) and (namespace._base_iri == old_base_iri):
+          del d[old_base_iri]
+          namespace._base_iri = new_base_iri
+          d[new_base_iri] = namespace
+          
+    else: # Create a Namespace for replacing the ontology
+      self._base_iri = new_base_iri
+      namespace = self.get_namespace(old_base_iri)
+      for entity in self.world._entities.values():
+        if entity.namespace is self: entity.namespace = namespace
+        
+    if self.world.graph: self.world.graph.release_write_lock()
+  base_iri = property(get_base_iri, set_base_iri)
+  
+  def destroy(self, update_relation = False, update_is_a = False):
     self.world.graph.acquire_write_lock()
-    del self.world.ontologies[self.base_iri]
+    
+    if update_relation:
+      for s, p in self.graph.execute("""SELECT DISTINCT s, p FROM quads WHERE c=?""", (self.graph.c,)):
+        entity = self.world._entities.get(s)
+        if (not entity is None) and (not entity.namespace.ontology is self):
+          prop = self.world._entities.get(p)
+          if not prop is None:
+            try: delattr(entity, prop.python_name)
+            except AttributeError: pass
+            
+    if update_is_a:
+      entities_needing_update = set()
+      for s, p in self.graph.execute("""SELECT DISTINCT s, o FROM objs WHERE c=? AND p=?""", (self.graph.c, rdf_type)):
+        entity = self.world._entities.get(s)
+        if (not entity is None) and (not entity.namespace.ontology is self): entities_needing_update.add(entity)
+        
+    del self.world.ontologies[self._base_iri]
     self.graph.destroy()
     for entity in list(self.world._entities.values()):
       if entity.namespace.ontology is self: del self.world._entities[entity.storid]
+      
+    if update_is_a:
+      for entity in entities_needing_update:
+        with LOADING:
+          entity.is_a = [self.world._get_by_storid(o) for o in self.world.graph._get_obj_triples_sp_o(entity.storid, rdf_type)]
+          
     self.world.graph.release_write_lock()
 
   def _entity_destroyed(self, entity): pass
@@ -860,8 +990,8 @@ class Ontology(Namespace, _GraphManager):
     if not r is None: return r
     return Namespace(self, base_iri, name or base_iri[:-1].rsplit("/", 1)[-1])
   
-  def __exit__(self, exc_type = None, exc_val = None, exc_tb = None):
-    Namespace.__exit__(self, exc_type, exc_val, exc_tb)
+  def __enter__(self): # Do this in __enter__ and not __exit__, because set_last_update_time() modify the database. Modifying the database in __exit__ prevents calling World.save() inside the 'with onto:' block. 
+    Namespace.__enter__(self)
     if not self.loaded:
       self.loaded = True
       if self.graph: self.graph.set_last_update_time(time.time())
@@ -875,14 +1005,17 @@ class Ontology(Namespace, _GraphManager):
         
   def load(self, only_local = False, fileobj = None, reload = False, reload_if_newer = False, url = None, **args):
     if self.loaded and (not reload): return self
-
-    if   self.base_iri in PREDEFINED_ONTOLOGIES:
-      f = os.path.join(os.path.dirname(__file__), "ontos", PREDEFINED_ONTOLOGIES[self.base_iri])
+    
+    if   self._base_iri in PREDEFINED_ONTOLOGIES:
+      f = os.path.join(os.path.dirname(__file__), "ontos", PREDEFINED_ONTOLOGIES[self._base_iri])
     elif not fileobj:
-      f = fileobj or _get_onto_file(self.base_iri, self.name, "r", only_local)
+      f = fileobj or _get_onto_file(self._base_iri, self.name, "r", only_local)
     else:
       f = ""
       
+    if reload_if_newer and not(f.startswith("http:") or f.startswith("https:")):
+      reload = os.path.getmtime(f) > self.graph.get_last_update_time()
+      
     self.world.graph.acquire_write_lock()
     
     if reload: self._destroy_cached_entities()
@@ -892,29 +1025,47 @@ class Ontology(Namespace, _GraphManager):
       if  reload or (self.graph.get_last_update_time() == 0.0): # Never loaded
         if _LOG_LEVEL: print("* Owlready2 *     ...loading ontology %s from %s..." % (self.name, f), file = sys.stderr)
         try:     fileobj = urllib.request.urlopen(url or f)
-        except:  raise OwlReadyOntologyParsingError("Cannot download '%s'!" % f)
-        try:     new_base_iri = self.graph.parse(fileobj, default_base = self.base_iri, **args)
+        except:  raise OwlReadyOntologyParsingError("Cannot download '%s'!" % (url or f))
+        #try:     new_base_iri = self.graph.parse(fileobj, default_base = self._base_iri, **args)
+        #finally: fileobj.close()
+        try:
+          new_base_iri = self.graph.parse(fileobj, default_base = self._base_iri, **args)
+        except OwlReadyOntologyParsingError:
+          if f.endswith(".owl") or f.endswith(".rdf") or f.endswith(".xml") or url: raise
+          else:
+            fileobj2 = None
+            for ext in ["owl", "rdf", "xml"]:
+              f2 = "%s.%s" % (f, ext)
+              try:
+                fileobj2 = urllib.request.urlopen(f2)
+                break
+              except: pass
+            if not fileobj2: raise
+            
+            try:     new_base_iri = self.graph.parse(fileobj2, default_base = self._base_iri, **args)
+            finally: fileobj2.close()
         finally: fileobj.close()
+        
     elif fileobj:
       if _LOG_LEVEL: print("* Owlready2 *     ...loading ontology %s from %s..." % (self.name, getattr(fileobj, "name", "") or getattr(fileobj, "url", "???")), file = sys.stderr)
-      try:     new_base_iri = self.graph.parse(fileobj, default_base = self.base_iri, **args)
+      try:     new_base_iri = self.graph.parse(fileobj, default_base = self._base_iri, **args)
       finally: fileobj.close()
     else:
-      if reload or (reload_if_newer and (os.path.getmtime(f) > self.graph.get_last_update_time())) or (self.graph.get_last_update_time() == 0.0):
+      #if reload or (reload_if_newer and (os.path.getmtime(f) > self.graph.get_last_update_time())) or (self.graph.get_last_update_time() == 0.0):
+      if reload or (self.graph.get_last_update_time() == 0.0):
         if _LOG_LEVEL: print("* Owlready2 *     ...loading ontology %s from %s..." % (self.name, f), file = sys.stderr)
         fileobj = open(f, "rb")
-        try:     new_base_iri = self.graph.parse(fileobj, default_base = self.base_iri, **args)
+        try:     new_base_iri = self.graph.parse(fileobj, default_base = self._base_iri, **args)
         finally: fileobj.close()
       else:
         if _LOG_LEVEL: print("* Owlready2 *     ...loading ontology %s (cached)..." % self.name, file = sys.stderr)
         
     self.loaded = True
 
-    if new_base_iri and (new_base_iri != self.base_iri):
-      self.graph.add_ontology_alias(new_base_iri, self.base_iri)
-      self.base_iri = new_base_iri
-      self._namespaces[self.base_iri] = self.world.ontologies[self.base_iri] = self
-      #if new_base_iri.endswith("#"):
+    if new_base_iri and (new_base_iri != self._base_iri):
+      self.graph.add_ontology_alias(new_base_iri, self._base_iri)
+      self._base_iri = new_base_iri
+      self._namespaces[self._base_iri] = self.world.ontologies[self._base_iri] = self
       if new_base_iri.endswith("#") or new_base_iri.endswith("/"):
         self.storid = self.world._abbreviate(new_base_iri[:-1])
       else:
@@ -988,7 +1139,7 @@ class Ontology(Namespace, _GraphManager):
       
   def save(self, file = None, format = "rdfxml", **kargs):
     if   file is None:
-      file = _open_onto_file(self.base_iri, self.name, "wb")
+      file = _open_onto_file(self._base_iri, self.name, "wb")
       if _LOG_LEVEL: print("* Owlready2 * Saving ontology %s to %s..." % (self.name, getattr(file, "name", "???")), file = sys.stderr)
       self.graph.save(file, format, **kargs)
       file.close()
@@ -1086,7 +1237,7 @@ class Ontology(Namespace, _GraphManager):
             self._del_obj_triple_spo  (bnode, None, None)
             self._del_data_triple_spod(bnode, None, None, None)
           return bnode
-
+        
         
   def _reload_bnode(self, bnode):
     if bnode in self._bnodes:
@@ -1208,7 +1359,7 @@ class Ontology(Namespace, _GraphManager):
       else:
         self._add_obj_triple_spo(bnode, rdf_rest, rdf_nil)
         
-  def __repr__(self): return """get_ontology("%s")""" % (self.base_iri)
+  def __repr__(self): return """get_ontology("%s")""" % (self._base_iri)
   
   def get_parents_of(self, entity):
     if isinstance(entity, Thing):
@@ -1234,6 +1385,11 @@ class Metadata(object):
     object.__setattr__(self, "namespace", namespace)
     object.__setattr__(self, "storid"   , storid)
     
+  def __iter__(self):
+    for p in self.namespace.world._get_triples_s_p(self.storid):
+      if p == rdf_type: continue
+      yield self.namespace.ontology._to_python(p)
+      
   def __getattr__(self, attr):
     Prop = self.namespace.world._props.get(attr)
     values = [self.namespace.ontology._to_python(o, d) for o, d in self.namespace.world._get_triples_sp_od(self.storid, Prop.storid)]
@@ -1260,7 +1416,7 @@ def _open_onto_file(base_iri, name, mode = "r", only_local = False):
   for dir in onto_path:
     for ext in ["", ".owl", ".rdf", ".n3"]:
       filename = os.path.join(dir, "%s%s" % (name, ext))
-      if os.path.exists(filename): return open(filename, mode)
+      if os.path.exists(filename) and os.path.isfile(filename): return open(filename, mode)
   if (mode.startswith("r")) and not only_local: return urllib.request.urlopen(base_iri)
   if (mode.startswith("w")): return open(os.path.join(onto_path[0], "%s.owl" % name), mode)
   raise FileNotFoundError
@@ -1271,10 +1427,10 @@ def _get_onto_file(base_iri, name, mode = "r", only_local = False):
   
   for dir in onto_path:
     filename = os.path.join(dir, base_iri.rsplit("/", 1)[-1])
-    if os.path.exists(filename): return filename
+    if os.path.exists(filename) and os.path.isfile(filename): return filename
     for ext in ["", ".nt", ".ntriples", ".rdf", ".owl"]:
       filename = os.path.join(dir, "%s%s" % (name, ext))
-      if os.path.exists(filename): return filename
+      if os.path.exists(filename) and os.path.isfile(filename): return filename
   if (mode.startswith("r")) and not only_local: return base_iri
   if (mode.startswith("w")): return os.path.join(onto_path[0], "%s.owl" % name)
   raise FileNotFoundError
