@@ -18,7 +18,7 @@
 
 
 import sys, os, re, urllib.parse, functools, datetime
-from owlready2.base import _universal_datatype_2_abbrev, _parse_datetime
+from owlready2.base import _universal_datatype_2_abbrev, _parse_datetime, rdf_type, owl_named_individual
 import owlready2, owlready2.rply as rply
 
 
@@ -79,6 +79,17 @@ _FUNC_2_DATATYPE = {
   #"REPLACE" :  # ok
   #"SIMPLEREPLACE" :  # ok
   "NEWINSTANCEIRI" : None, # ok
+  "LOADED" : _universal_datatype_2_abbrev[bool], # ok
+  "STORID" : _universal_datatype_2_abbrev[int], # ok
+  "DATE"    : _universal_datatype_2_abbrev[datetime.date], # ok
+  "TIME"    : _universal_datatype_2_abbrev[datetime.time], # ok
+  "DATETIME": _universal_datatype_2_abbrev[datetime.datetime], # ok
+  "DATE_DIFF": _universal_datatype_2_abbrev[datetime.timedelta], # ok
+  "DATE_ADD": _universal_datatype_2_abbrev[datetime.date], # ok
+  "DATE_SUB": _universal_datatype_2_abbrev[datetime.date], # ok
+  "DATETIME_DIFF": _universal_datatype_2_abbrev[datetime.timedelta], # ok
+  "DATETIME_ADD": _universal_datatype_2_abbrev[datetime.datetime], # ok
+  "DATETIME_SUB": _universal_datatype_2_abbrev[datetime.datetime], # ok
 
   "HTTP://WWW.W3.ORG/2001/XMLSCHEMA#INTEGER" : _universal_datatype_2_abbrev[int], # ok
   "HTTP://WWW.W3.ORG/2001/XMLSCHEMA#DOUBLE" : _universal_datatype_2_abbrev[float], # ok
@@ -159,7 +170,26 @@ def _timezone(x):
     "%dM" % minutes if minutes else  "",
     "%dS" % delta.seconds if (not days and not hours and not minutes) else "",
   )
-  
+
+def _date_diff(d1, d2):
+  return owlready2.base._format_duration(datetime.date.fromisoformat(d1) - datetime.date.fromisoformat(d2))
+
+def _date_add(d, td):
+  return (datetime.date.fromisoformat(d) + owlready2.base._parse_duration(td)).isoformat()
+
+def _date_sub(d, td):
+  return (datetime.date.fromisoformat(d) - owlready2.base._parse_duration(td)).isoformat()
+
+def _datetime_diff(d1, d2):
+  return owlready2.base._format_duration(abs(datetime.datetime.fromisoformat(d1) - datetime.datetime.fromisoformat(d2)))
+
+def _datetime_add(d, td):
+  return (datetime.datetime.fromisoformat(d) + owlready2.base._parse_duration(td)).isoformat()
+
+def _datetime_sub(d, td):
+  return (datetime.datetime.fromisoformat(d) - owlready2.base._parse_duration(td)).isoformat()
+
+
 class _Func(object):
   def __init__(self, world):
     self.world        = world
@@ -188,13 +218,19 @@ class _Func(object):
     return bn
   
   def _newinstanceiri(self, x):
-    x = self.world._get_by_storid(x)
-    namespace = (owlready2.CURRENT_NAMESPACES.get() and owlready2.CURRENT_NAMESPACES.get()[-1]) or x.namespace
-    iri = self.world.graph._new_numbered_iri("%s%s" % (namespace.base_iri, x.name.lower()))
-    return self.world._abbreviate(iri)
+    Class = self.world._get_by_storid(x)
+    namespace = (owlready2.CURRENT_NAMESPACES.get() and owlready2.CURRENT_NAMESPACES.get()[-1]) or Class.namespace
+    iri = self.world.graph._new_numbered_iri("%s%s" % (namespace.base_iri, Class.name.lower()))
+    storid = self.world._abbreviate(iri)
+    namespace.ontology._add_obj_triple_spo(storid, rdf_type, owl_named_individual)
+    namespace.ontology._add_obj_triple_spo(storid, rdf_type, x)
+    return storid
   
+  def _loaded(self, x):
+    return x in self.world._entities
   
-def register_python_function(world):
+
+def register_python_builtin_functions(world):
   if (sys.version_info.major == 3) and (sys.version_info.minor < 8):
     def create_function(name, num_params, func, deterministic = False):
       world.graph.db.create_function(name, num_params, func)
@@ -208,6 +244,12 @@ def register_python_function(world):
   create_function("seconds",        1, _seconds,  deterministic = True)
   create_function("tz",             1, _tz,       deterministic = True)
   create_function("timezone",       1, _timezone, deterministic = True)
+  create_function("date_diff",      2, _date_diff, deterministic = True)
+  create_function("date_add",       2, _date_add, deterministic = True)
+  create_function("date_sub",       2, _date_sub, deterministic = True)
+  create_function("datetime_diff",  2, _datetime_diff, deterministic = True)
+  create_function("datetime_add",   2, _datetime_add, deterministic = True)
+  create_function("datetime_sub",   2, _datetime_sub, deterministic = True)
   create_function("encode_for_uri", 1, urllib.parse.quote, deterministic = True)
   create_function("uuid",           0, _uuid)
   create_function("struuid",        0, _struuid)
@@ -219,12 +261,12 @@ def register_python_function(world):
   create_function("now",             0, func._now, deterministic = True)
   create_function("bnode",          -1, func._bnode)
   create_function("newinstanceiri",  1, func._newinstanceiri)
+  create_function("loaded",          1, func._loaded)
   
   # Unindexed table for deprioritizing subqueries
-  world.graph.execute("""CREATE TEMP TABLE one (i INTEGER)""") # CREATE TEMP TABLE one (i INTEGER); INSERT INTO one VALUES (1);
+  world.graph.execute("""CREATE TEMP TABLE one (i INTEGER)""")
   world.graph.execute("""INSERT INTO one VALUES (1)""")
-  
-  
+
 class FuncSupport(object):
   def parse_expression(self, expression):
     if   hasattr(expression, "sql"):   return " %s" % expression.sql
@@ -235,7 +277,12 @@ class FuncSupport(object):
           if   func == "CONTAINS":
             return "(INSTR(%s)!=0)" % self.parse_expression(expression[2])
           elif func == "STRSTARTS":
-            return "(INSTR(%s)=1)" % self.parse_expression(expression[2])
+            x     = self.parse_expression(expression[2][0])
+            start = self.parse_expression(expression[2][2]).strip()
+            if start.startswith("'") and start.endswith("'") and not "'" in start[1:-1]:
+              return "(SUBSTR(%s,1,%s)=%s)" % (x, len(start) - 2, start)
+            else:
+              return "(INSTR(%s,%s)=1)" % (x, start)
           elif func == "STRENDS":
             eo1 = self.parse_expression(expression[2][0])
             eo2 = self.parse_expression(expression[2][2])
@@ -335,6 +382,8 @@ class FuncSupport(object):
             return self.parse_expression(expression[2][0])
           elif func == "SAMPLE":
             return "MIN(%s)" % self.parse_expression(expression[3])
+          elif func == "STORID":
+            return self.parse_expression(expression[2][0])
         elif isinstance(expression[0], rply.Token) and expression[0].name == "ASC":
           return "%s ASC" % self.parse_expression(expression[1][1])
         elif isinstance(expression[0], rply.Token) and expression[0].name == "DESC":
